"""
Helper functions/classes for ghdorker
"""
from datetime import datetime
from time import sleep
import logging

logger = logging.getLogger("debug")

class Limit():
  """A class for handling single endpoint limits

  Parameters:
  limit (int): Max number of requests
  remaining (int): Number of remaining allowable calls
  reset (datetime): When the limit resets
  used (int): Number of calls used so far
  resource (str): Name of the endpoint
  """
  def __init__(self, limit, remaining, reset, used, resource):
    self.limit = limit
    self.remaining = remaining
    self.reset = reset
    self.used = used
    self.resource = resource

  def step(self):
    """Iterate the counters"""
    self.remaining = self.remaining -1
    self.used = self.used + 1

  def __str__(self) -> str:
    return f"""\n
            limit: {self.limit}
            remaining: {self.remaining}
            reset: {self.reset}
            used: {self.used}
            resource: {self.resource}"""


class RateLimiter:
  """A class for ensuring rate limits are not exceeded

  Parameters:
  client (GHApi): A GHApi instance
  """
  def __init__(self, client):
    self.client = client
    self.initialize_limits()


  def get_rate_limits(self, endpoint: str) -> str:
    """Get the currently set rate limits for an endpoint"""
    rate_limit = getattr(self, endpoint)
    return str(rate_limit)


  def initialize_limits(self):
    """Populate limits from GitHub

    This will auto populate the class with limits objects
    based on the endpoints in GitHub. This way it can get
    real limits based off your currently used limits.s"""
    limits = self.client.rate_limit.get()
    for key, value in limits["resources"].items():
      setattr(self, key, Limit(**value))


  def check_safety(self, endpoint: str):
    """Checks the safety of making a call to github

    Return:
    There is no return value but it will pause for a short period to
    prevent overusing the endpoints."""
    rate_limit = getattr(self, endpoint)
    curr_time = datetime.now().timestamp()

    if curr_time < rate_limit.reset:
      sleep(((rate_limit.reset - curr_time) / rate_limit.remaining)+1)
    else:
      self.initialize_limits()

    rate_limit.step()


def paginator(operation, per_page=30, page=1, **kwargs):
  """Helper function for paginating requests

  Parameters:
  operation (GHapi Function): The fuction you would like to paginate requests from
  per_page (int): Number of results per page (some endpoints have a 30 max)
  page (int): Page to start on
  kwargs: any other arguments you would like to pass to the funtion (eg. q=Query)

  Returns:
  Attribute Dict: A list of dictionary objects containing the results returned
  """
  incomplete = True
  while incomplete:
    result = operation(**kwargs, per_page=per_page, page=page)
    incomplete = result['incomplete_results']
    yield result
    page += 1



import logging
import json
import argparse
from os import getenv
from dotenv import load_dotenv
from ghapi.all import GhApi
from GHDorker.helpers import RateLimiter
from GHDorker.helpers import paginator

load_dotenv()
GH_TOKEN = getenv('GH_TOKEN', None)

#### Logging config
console_out = logging.getLogger("ghdorker")
consoleOutHandle = logging.StreamHandler()
consoleOutHandle.setLevel(logging.INFO)
consoleOutFormatter = logging.Formatter('%(asctime)s - %(message)s')
consoleOutHandle.setFormatter(consoleOutFormatter)
console_out.addHandler(consoleOutHandle)
console_out.setLevel(logging.INFO)

logger = logging.getLogger("debug")
consoleHandle = logging.StreamHandler()
consoleHandle.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
consoleHandle.setFormatter(formatter)
logger.addHandler(consoleHandle)
logger.setLevel(logging.ERROR)


def get_client() -> GhApi:
  """Return the GitHub Client"""
  logger.debug("Creating GitHub API Object %s GitHub Token",
               'with' if GH_TOKEN is not None else 'without')

  return GhApi(token=GH_TOKEN)


def search_results(query: str, client: GhApi):
  """Yeilds results pages"""
  search_gen = paginator(client.search.code, q=query)
  rate_limits = RateLimiter(client)

  for results in search_gen:
    logger.debug("Current Rate Limits: %s",
                 rate_limits.get_rate_limits('search'))
    rate_limits.check_safety("search")
    yield results


def input_file_parse(file_path: str) -> str:
  """Parse the dorkfile"""
  logger.debug("Opening %s", file_path)

  with open(file_path, 'r', encoding="UTF-8") as infile:
    logger.debug("Reading %s", file_path)
    return [line.strip() for line in infile.readlines()]


def output_json_file(data, filename):
  """Output the data as a JSON File"""
  if len(data) == 0:
    return

  with open(filename, "+w", encoding="UTF-8") as outfile:
    for entry in data:
      json.dump(entry, outfile)
      outfile.write("\n")


def output_csv_file(data, filename):
  """Output the data as a CSV File"""
  if len(data) == 0:
    return

  logger.debug("Writing CSV File: %s", filename)
  with open(filename, "+w", encoding="UTF-8") as outfile:
    outfile.write("dork, repository, path, score\n")
    for entry in data:
      outfile.write(f"{entry['dork']}, {entry['repository']}, {entry['path']}, {entry['score']}\n")


def output_format(results):
  """Formatter for data"""
  logger.debug("Formatting Results")
  return [
    {
      "dork": item["dork"],
      "repository": item["repository"]["full_name"],
      "path": item["path"],
      "score": item["score"]
    }
    for item in results
  ]


def console_log_ouput(item):
  """Function for logging results to console"""
  console_out.info(f"dork: {item['dork']}, "
                   f"repository: {item['repository']['full_name']}, "
                   f"path: {item['path']}, "
                   f"score: {item['score']}")


def main(dorks, scope, search, output_filename, debug):
  """Main logic of the program"""
  if debug:
    logger.setLevel(logging.DEBUG)

  dork_list = input_file_parse(dorks)
  query_list = [f"{dork} {scope}:{search}" for dork in dork_list]
  client = get_client()

  results = []

  for query in query_list:
    for result in search_results(query, client):
      updated_results = [{"dork": query, **item} for item in result["items"]]

      map(console_log_ouput, updated_results)

      results.extend(updated_results)

  formatted_results = output_format(results)

  if output_filename.endswith('.json'):
    output_json_file(formatted_results, output_filename)
  elif output_filename.endswith('.csv'):
    output_csv_file(formatted_results, output_filename)


def cli_entry():
  """Parse arguments and kickoff the process"""
  parser = argparse.ArgumentParser(
    description='Search github for github dorks',
    epilog='Use responsibly, Enjoy pentesting')

  parser.add_argument(
    '-v',
    '--version',
    action='version',
    version='%(prog)s 0.2.0')

  parser.add_argument(
    '-s',
    '--scope',
    choices=['repo', 'user'],
    help='The type of GitHub object you would like to search')

  parser.add_argument(
    '-d',
    '--dorks',
    default='github-dorks.txt',
    help='Github dorks file. Eg: github-dorks.txt')


  parser.add_argument(
    '--debug',
    action='store_true',
    help='Set this if you would like to see verbose logging.')


  parser.add_argument(
    '-o',
    '--outputFile',
    dest='output_filename',
    action='store',
    help="""File to write results to. This overwrites the file provided!\n
            Accepts .json or .csv as output file types.""")

  parser.add_argument(
    'search',
    help='The GitHub object you would like to search (eg. repo or username)')

  args = parser.parse_args()
  main(**vars(args))


if __name__=='__main__':
  cli_entry()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="ghdorker",
    version='0.2.0',
    author="David Tippett",
    author_email="dtaivpp@gmail.com",
    description="A better GitHub Dorking Utility",
    license='Apache Software License',
    license_files=['LICENSE'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="GitHub Dorker Python Security Dorking",
    url="https://github.com/dtaivpp/gh-dorker",
    packages=['GHDorker'],
    install_requires=['ghapi','python-dotenv', 'wheel'],
    entry_points = {
        'console_scripts': ['ghdorker=GHDorker.dorker:cli_entry'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        'Intended Audience :: Developers',
        "Operating System :: OS Independent"
    ],
)
